import os
import boto3
from boto3.session import Session
import configparser
import argparse
import sys
from typing import Any, Tuple, List

from app.dynamodb import csv_import, csv_export, truncate, move

__version__ = "1.5.0"
config_file = "config.ini"


def main() -> None:
    """Main routine
    """
    (message, code) = execute()
    print(message)
    sys.exit(code)


def execute() -> Tuple:
    """Command execute

    Raises:
        ValueError: invalied config

    Returns:
        Tuple: result message and exit code
    """

    result = "No operations."

    # arguments parse
    args = parse_args(sys.argv[1:])

    # boto3 config setting
    try:
        tables = config_read_and_get_table(args)
    except ValueError as e:
        return (str(e), 1)

    except Exception:
        return (f"Invalid format {config_file} file", 1)

    # csv import
    if args.imp:
        if args.file is not None:
            result = csv_import(tables[0], args.file, args.ignore)
        else:
            return ("Import mode requires a input file option.", 1)

    # csv export
    if args.exp:
        if args.output is not None:
            parameters = {}
            if args.index is not None:
                parameters["IndexName"] = args.index
            result = csv_export(tables[0], args.output, parameters)
        else:
            return ("Export mode requires a output file option.", 1)

    # truncate table
    if args.truncate:
        result = truncate(tables[0])

    # move table
    if args.move:
        if len(tables) == 2:
            result = move(tables)
        else:
            return ("Move mode requires a two tables.", 1)

    return result


def parse_args(args: str) -> Any:
    """Parse arguments

    Args:
        args (str): _description_

    Returns:
        Any: parsed args
    """
    parser = argparse.ArgumentParser(
        description="Import CSV file into DynamoDB table utilities")
    parser.add_argument("-v", "--version", action="version",
                        version=__version__,
                        help="show version")
    parser.add_argument(
        "-i", "--imp", help="mode import", action="store_true")
    parser.add_argument(
        "-e", "--exp", help="mode export", action="store_true")
    parser.add_argument(
        "--truncate", help="mode truncate", action="store_true")
    parser.add_argument(
        "--move", help="mode move", action="store_true")
    parser.add_argument(
        "-t", "--table", nargs="*", help="DynamoDB table name", required=True)
    parser.add_argument(
        "-idx", "--index", help="DynamoDB index name")
    parser.add_argument(
        "-f", "--file", help="UTF-8 CSV file path required import mode")
    parser.add_argument(
        "-o", "--output", help="output file path required export mode")
    parser.add_argument(
        "--ignore", help="ignore import error", action="store_true")
    parser.add_argument(
        "--profile", help="using AWS profile")

    return parser.parse_args()


def config_read_and_get_table(args: Any) -> List:
    """Config read and Create DynamoDB table instance

    Args:
        args (Any): arguments

    Returns:
        List: DynamoDB tables class
    """
    # using AWS profile
    if args.profile:
        session = Session(profile_name=args.profile)
        dynamodb = session.resource('dynamodb')
    else:
        # using config
        if not os.path.isfile(config_file):
            raise ValueError(f"Please make your {config_file} file")

        config = configparser.ConfigParser()
        config.optionxform = str
        config.read_dict({"AWS": {"ENDPOINT_URL": ""}})
        config.read(config_file)

        endpoint_url = None
        if config.get("AWS", "ENDPOINT_URL"):
            endpoint_url = config.get("AWS", "ENDPOINT_URL")
        dynamodb = boto3.resource("dynamodb",
            region_name=config.get("AWS", "REGION"),
            aws_access_key_id=config.get("AWS", "AWS_ACCESS_KEY_ID"),
            aws_secret_access_key=config.get("AWS", "AWS_SECRET_ACCESS_KEY"),
            endpoint_url=endpoint_url)

    tables = []
    for table in args.table:
        tables.append(dynamodb.Table(table))

    return tables


if __name__ == "__main__":
    main()



from botocore.exceptions import ClientError
from tqdm import tqdm
from typing import Tuple, List


def move(tables: List) -> Tuple:
    """DynamoDB table items move

    Args:
        tables (List): boto3 DynamoDB table objects

    Returns:
        Tuple: result message and exit code
    """

    from_table = tables[0]
    to_table = tables[1]

    # all scan move items
    move_items = []
    parameters = {}

    try:
        while True:
            response = from_table.scan(**parameters)
            move_items.extend(response["Items"])
            if ("LastEvaluatedKey" in response):
                parameters["ExclusiveStartKey"] = response["LastEvaluatedKey"]
            else:
                break
    except ClientError as e:
        return (f"aws client error:{e}", 1)
    except Exception as e:
        return (f"table not found:{e}", 1)

    print("{name} scan {count} items".format(
        name=from_table.name, count=len(move_items)))

    # move all items
    try:
        with to_table.batch_writer() as batch:
            print("please wait {name} moving".format(name=to_table.name))

            for item in tqdm(move_items):
                batch.put_item(Item=item)

    except ClientError as e:
        return (f"aws client error:{e}", 1)
    except Exception as e:
        return (f"table not found:{e}", 1)

    return ("{name} moved {count} items".format(
        name=to_table.name, count=len(move_items)), 0)



import configparser
from boto3.dynamodb.conditions import Key
from botocore.exceptions import ClientError
from tqdm import tqdm
import csv
import json
from decimal import Decimal
from typing import Any, Tuple, Dict


def csv_export(table: Any, file: str, parameters: Dict = {}) -> Tuple:
    """Export DynamoDB table to csv

    Args:
        table (Table): boto3 DynamoDB table object
        file (str): csv file path

    Returns:
        Tuple: result message and exit code
    """

    # read csv spec
    try:
        csv_spec = configparser.ConfigParser()
        csv_spec.optionxform = str
        csv_spec.read(f"{file}.spec")
    except Exception as e:
        return (f"CSV specification file can't read:{e}", 1)

    # write csv
    try:
        with open(file, mode="w", encoding="utf_8") as f:
            print("please wait {name} exporting {file}".format(
                name=table.name, file=file))

            export_items = []
            try:
                if "QUERY_OPTION" in csv_spec:
                    try:
                        # Partition key option
                        if "PKAttribute" in csv_spec["QUERY_OPTION"]:
                            pk_key = csv_spec.get("QUERY_OPTION", "PKAttribute")
                            pk_value = csv_spec.get("QUERY_OPTION", "PKAttributeValue")
                            pk_type = csv_spec.get("QUERY_OPTION", "PKAttributeType")
                            if pk_type == "I":
                                pk_value = int(pk_value)
                            parameters["KeyConditionExpression"] = Key(pk_key).eq(pk_value)

                        # Sort key option
                        if "SKAttribute" in csv_spec["QUERY_OPTION"]:
                            sk_key = csv_spec.get("QUERY_OPTION", "SKAttribute")
                            sk_values = csv_spec.get("QUERY_OPTION", "SKAttributeValues").split(",")
                            sk_type = csv_spec.get("QUERY_OPTION", "SKAttributeType")

                            if sk_type == "I":
                                sk_values = [int(v) for v in sk_values]

                            sk_values = sk_values[0] if len(sk_values) == 1 else sk_values

                            sk_expression = csv_spec.get("QUERY_OPTION", "SKAttributeExpression")
                            if sk_expression == "begins_with":
                                parameters["KeyConditionExpression"] &= Key(sk_key).begins_with(sk_values)
                            elif sk_expression == "between":
                                parameters["KeyConditionExpression"] &= Key(sk_key).between(sk_values[0], sk_values[1])
                            elif sk_expression == "eq":
                                parameters["KeyConditionExpression"] &= Key(sk_key).eq(sk_values)
                            elif sk_expression == "gt":
                                parameters["KeyConditionExpression"] &= Key(sk_key).gt(sk_values)
                            elif sk_expression == "gte":
                                parameters["KeyConditionExpression"] &= Key(sk_key).gte(sk_values)
                            elif sk_expression == "lt":
                                parameters["KeyConditionExpression"] &= Key(sk_key).lt(sk_values)
                            elif sk_expression == "lte":
                                parameters["KeyConditionExpression"] &= Key(sk_key).lte(sk_values)

                        # query table
                        while True:
                            response = table.query(**parameters)
                            export_items.extend(response["Items"])
                            if ("LastEvaluatedKey" in response):
                                parameters["ExclusiveStartKey"] = response["LastEvaluatedKey"]
                            else:
                                break
                    except Exception as e:
                        return (f"query option error:{e}", 1)

                else:
                    # scan table
                    while True:
                        response = table.scan(**parameters)
                        export_items.extend(response["Items"])
                        if ("LastEvaluatedKey" in response):
                            parameters["ExclusiveStartKey"] = response["LastEvaluatedKey"]
                        else:
                            break
            except ClientError as e:
                return (f"aws client error:{e}", 1)
            except Exception as e:
                return (f"table not found:{e}", 1)

            is_write_csv_header_labels = False
            for item in tqdm(export_items):

                if not is_write_csv_header_labels:
                    # write csv header labels
                    csv_header_labels = list(csv_spec["CSV_SPEC"])
                    writer = csv.DictWriter(f, fieldnames=csv_header_labels, lineterminator="\n")
                    writer.writeheader()
                    is_write_csv_header_labels = True

                # updated dict to match specifications
                for key in list(item.keys()):
                    try:
                        spec = csv_spec.get("CSV_SPEC", key)
                    except Exception:
                        # Removed attributes that do not match the specifications
                        del item[key]
                        continue

                    item[key] = convert_item(spec, item, key)

                writer.writerow(item)

        return ("{name} csv exported {count} items".format(
            name=table.name, count=len(export_items)), 0)

    except IOError as e:
        print(f"I/O error:{e}")

    except Exception as e:
        return (str(e), 1)


def convert_item(spec: str, item: Dict, key: str) -> Any:
    """convert item

    Args:
        spec (str): type of item
        row (Dict): row data
        key (str): key

    Returns:
        Any: converted item value
    """
    if spec == "S":  # String
        return str(item[key])
    elif spec == "I":  # Integer
        return int(item[key])
    elif spec == "D":  # Decimal
        return float(item[key])
    elif spec == "B":  # Boolean
        if not item[key]:
            return ""
    elif spec == "J":  # Json
        return json.dumps(item[key], default=decimal_encode)
    elif spec == "SL" or spec == "SS":  # StringList or StringSet
        return " ".join(item[key])
    elif spec == "DL" or spec == "DS":  # DecimalList or DecimalSetDecimalList
        return " ".join(list(map(str, item[key])))
    else:
        return item[key]


def decimal_encode(obj: Any) -> float:
    """encode decimal

    Args:
        obj (Any): object

    Raises:
        TypeError: Not decimal object

    Returns:
        float: decimal object
    """
    if isinstance(obj, Decimal):
        return float(obj)
    raise TypeError



import configparser
from tqdm import tqdm
import csv
import json
from decimal import Decimal
from typing import Any, Dict, Tuple

count = 0
error_count = 0


def csv_import(table: Any, file: str, ignore: bool = False) -> Tuple:
    """csv import into DynamoDB table

    Args:
        table (Any): boto3 DynamoDB table object
        file (str): csv file path
        ignore (bool): ignore put item error

    Returns:
        Tuple: result message and exit code
    """

    # read csv spec
    try:
        csv_spec = configparser.ConfigParser()
        csv_spec.optionxform = str
        csv_spec.read(f"{file}.spec")
    except Exception as e:
        return (f"CSV specification file can't read:{e}", 1)

    # read csv
    try:
        with open(file, mode="r", encoding="utf_8") as f:
            reader = csv.DictReader(f)

            # batch_size = 100
            batch = []

            print("please wait {name} importing {file}".format(
                name=table.name, file=file))
            for row in tqdm(reader):
                # No use buffer
                # if len(batch) >= batch_size:
                #     write_to_dynamo(table, batch)
                #     batch.clear()

                # updated dict to match specifications
                for key in list(row.keys()):
                    spec = csv_spec.get("CSV_SPEC", key)

                    # Convert blank value
                    if "IMPORT_OPTION" in csv_spec:
                        if "ConvertBlankToNullAttrs" in csv_spec["IMPORT_OPTION"] and not row[key]:
                            to_null_attrs = csv_spec.get("IMPORT_OPTION", "ConvertBlankToNullAttrs").split(",")
                            if key in to_null_attrs:
                                row[key] = None
                                continue

                        if "ConvertBlankToDropAttrs" in csv_spec["IMPORT_OPTION"] and not row[key]:
                            to_drop_attrs = csv_spec.get("IMPORT_OPTION", "ConvertBlankToDropAttrs").split(",")
                            if key in to_drop_attrs:
                                del row[key]
                                continue

                    try:
                        row[key] = convert_column(spec, row, key)
                    except Exception:
                        del row[key]

                batch.append(row)

            if (len(batch)) > 0:
                write_to_dynamo(table, batch, ignore)

        if ignore:
            message = "{name} csv imported {count} items and {error_count} error items".format(
                name=table.name, count=count, error_count=error_count)
        else:
            message = "{name} csv imported {count} items".format(
                name=table.name, count=count)
        return (message, 0)

    except Exception as e:
        return (f"CSV file can't read:{e}", 1)


def convert_column(spec: str, row: Dict, key: str) -> Any:
    """convert column

    Args:
        spec (str): type of column
        row (Dict): row data
        key (str): key

    Returns:
        Any: converted column value
    """
    if spec == "S":  # String
        return str(row[key])
    elif spec == "I":  # Integer
        return int(row[key])
    elif spec == "D":  # Decimal
        return Decimal(row[key])
    elif spec == "B":  # Boolean
        return bool(row[key])
    elif spec == "J":  # Json
        return json.loads(row[key], parse_float=Decimal)
    elif spec == "SL":  # StringList
        return row[key].split()
    elif spec == "SS":  # StringSet
        return set(row[key].split())
    elif spec == "DL":  # DecimalList
        return list(map(Decimal, row[key].split()))
    elif spec == "DS":  # DecimalSet
        return set(list(map(Decimal, row[key].split())))
    else:
        return row[key]


def write_to_dynamo(table: Any, rows: Dict, ignore: bool = False) -> None:
    """csv rows into DynamoDB

    Args:
        table (Any): boto3 DynamoDB table object
        rows (Dict): csv rows
        ignore (bool): ignore put item error
    """
    global count
    global error_count

    # Ignore error item
    if ignore:
        for i in tqdm(range(len(rows))):
            try:
                table.put_item(
                    Item=rows[i]
                )
                count = count + 1
            except Exception:
                error_count = error_count + 1

    # Batch write
    else:
        try:
            # overwrite duplicate key item
            key_names = [x["AttributeName"] for x in table.key_schema]
            with table.batch_writer(overwrite_by_pkeys=key_names) as batch:
                for i in tqdm(range(len(rows))):
                    batch.put_item(
                        Item=rows[i]
                    )
                    count = count + 1

        except Exception as e:
            print(f"Error executing batch_writer:{e}")



from botocore.exceptions import ClientError
from tqdm import tqdm
from typing import Any, Tuple


def truncate(table: Any) -> Tuple:
    """DynamoDB table truncate

    Args:
        table (Any): boto3 DynamoDB table object

    Returns:
        Tuple: result message and exit code
    """

    # all scan delete items
    delete_items = []
    parameters = {}

    try:
        while True:
            response = table.scan(**parameters)
            delete_items.extend(response["Items"])
            if ("LastEvaluatedKey" in response):
                parameters["ExclusiveStartKey"] = response["LastEvaluatedKey"]
            else:
                break
    except ClientError as e:
        return (f"aws client error:{e}", 1)
    except Exception as e:
        return (f"table not found:{e}", 1)

    print("{name} scan {count} items".format(
        name=table.name, count=len(delete_items)))

    # get key attribute name and item delete keys
    key_names = [x["AttributeName"] for x in table.key_schema]
    delete_keys = [{k: v for k, v in x.items() if k in key_names}
                for x in delete_items]

    # delete all items
    with table.batch_writer() as batch:
        print("please wait {name} truncating".format(name=table.name))

        for key in tqdm(delete_keys):
            batch.delete_item(Key=key)

    return ("{name} truncated".format(name=table.name), 0)
